import abc
import queue
import warnings
from dataclasses import dataclass, field
from enum import Enum
from threading import Thread
from typing import Any, Optional, Iterable, Callable

import numpy as np
import torch
import wandb
from wandb.sdk.data_types.base_types.media import Media

from ncalib.seed_factory import NCA_SeedFactory
from ncalib.trainer import NCA_Trainer, NCA_StoppingCriterion
from ncalib.trainer.batcher.simple import SimpleBatcher
from ncalib.trainer.dataset import DataSample, NCA_Dataset
from ncalib.trainer.logger.wandb import WandBLogger
from ncalib.trainer.loss_function import LossResult, NCA_LossFunction
from ncalib.visualization import np_make_grid, states_to_rgb_pillow, generate_video_array, blend_alpha, \
    _legacy_generate_video_array, as_quadratic_shape_as_possible, video_timeline_preparation, run_inference
from ncalib.visualization.state_visualizer import StateVisualizer
from ncalib.visualization.state_visualizer.simple import RGBVisualizer

WANDB_LOG_QUEUE: Optional[queue.Queue] = None


class WandbPlaceholder(abc.ABC):
    def replace(self) -> Media | list[Media]:
        pass


class WandBImagePlaceholder(WandbPlaceholder):
    def __init__(self, image_data: np.ndarray, caption: Optional[str] = None, **kwargs):
        self.image_data = image_data
        self.caption = caption
        self.kwargs = kwargs

    def replace(self) -> Media:
        return wandb.Image(self.image_data, caption=self.caption, **self.kwargs)


class WandBImageListPlaceholder(WandbPlaceholder):
    def __init__(self, images_data: list[np.ndarray], **kwargs):
        self.images_data = images_data
        self.kwargs = kwargs

    def replace(self) -> list[Media]:
        return [wandb.Image(image_data, **self.kwargs) for image_data in self.images_data]


class WandBVideoPlaceholder(WandbPlaceholder):
    def __init__(self, video_data: np.ndarray, fps: int, format: str = "gif", caption: Optional[str] = None):
        self.video_data = video_data
        self.fps = fps
        self.format = format
        self.caption = caption

    def replace(self) -> wandb.Video:
        return wandb.Video(self.video_data, caption=self.caption, fps=self.fps, format=self.format)


def handle_wandb_logging_queue():
    step = 0
    log_data = {}

    def log_now():
        nonlocal log_data
        if len(log_data) > 0:
            wandb.log(
                log_data,
                step=step,
            )
            log_data = {}

    while True:
        try:
            item = WANDB_LOG_QUEUE.get(timeout=2)
        except queue.Empty:
            log_now()
            continue

        # Escape WandB Log Queue
        if item is False:
            break

        this_step, this_log_data = item
        if this_step != step:
            # Log all aggregated data
            log_now()

            # Reset for next step
            step = this_step

        # Add new data
        for name in this_log_data:
            value = this_log_data[name]
            if isinstance(value, WandbPlaceholder):
                value = value.replace()
            log_data[name] = value

            if len(log_data) >= 10:
                log_now()  # Clear buffer

    log_now()


class WandBMetricLogger(WandBLogger, abc.ABC):
    """
    Creates an interface which has to implement a get_log_data function that returns a dict[name -> Any] that will
    get logged for the current epoch.
    """

    def __init__(
            self,
            *,
            epoch_interval: int = 1000,
            log_final: bool = False,
            prefix: Optional[str] = None
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final)
        self.prefix = prefix
        global WANDB_LOG_QUEUE
        self.wandb_log_queue_thread = None
        if WANDB_LOG_QUEUE is None:  # Make sure the wandb log queue is only initialized once
            WANDB_LOG_QUEUE = queue.Queue()
            self.wandb_log_queue_thread = Thread(target=handle_wandb_logging_queue, daemon=True)
            self.wandb_log_queue_thread.start()

    def __del__(self):
        if WANDB_LOG_QUEUE is not None:
            WANDB_LOG_QUEUE.put(False)

    def _log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        log_data = self.get_log_data(data_sample, state_progression, loss)
        if self.prefix is not None:
            log_data = {f"{self.prefix}/{k}": v for k, v in log_data.items()}
        WANDB_LOG_QUEUE.put((self.trainer.epoch, log_data))

    @abc.abstractmethod
    def get_log_data(
            self,
            data_sample: DataSample,
            state_progression: torch.Tensor,
            loss: LossResult
    ) -> dict[str, Any]:
        pass


class WandBLossLogger(WandBMetricLogger):
    def init_training(self, trainer: NCA_Trainer, stopping_criterion: NCA_StoppingCriterion):
        super().init_training(trainer, stopping_criterion)

    def get_log_data(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        detailed_loss = loss.detailed_loss(include_total=True)
        log_data = {f"loss/{name}": loss for name, (loss, weight) in detailed_loss.items()}
        log_data.update({f"loss_weighted/{name}": loss * weight for name, (loss, weight) in detailed_loss.items() if
                         name != "total"})

        return log_data


class BatchMode(Enum):
    MERGED_QUADRATIC = 1
    MERGED_ROW = 2
    MERGED_COLUMN = 3
    SEPARATE = 4


class WandBBatchLogger(WandBMetricLogger):
    def __init__(
            self,
            visualizer: StateVisualizer = RGBVisualizer(),
            *,
            log_name: str = "batch",
            epoch_interval: int = 100,
            log_final: bool = False,
            target_visualizer: bool | StateVisualizer = True,
            batch_mode: BatchMode = BatchMode.MERGED_COLUMN,
            prefix: Optional[str] = None,
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final, prefix=prefix)
        self.batch_mode = batch_mode
        self.log_name = log_name
        self.visualizer = visualizer
        if target_visualizer is True:
            target_visualizer = visualizer
        self.target_visualizer: StateVisualizer | bool = target_visualizer

    def get_log_data(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        log_data = {}
        log_data.update(self._create_visualizations(state_progression[-1], "state", visualizer=self.visualizer))
        log_data.update(self._create_visualizations(state_progression[0], "seeds", visualizer=self.visualizer))
        log_data.update(self._create_visualizations(data_sample.targets, "targets", visualizer=self.target_visualizer))

        return log_data

    def _create_visualizations(
            self,
            state: torch.Tensor,
            name: str,
            visualizer: StateVisualizer | bool
    ) -> dict[str, WandBImagePlaceholder | WandBImageListPlaceholder]:
        if visualizer is False:
            return {}
        result = {}
        visualization = visualizer(state)

        for visualization_name, images in visualization.items():
            if self.batch_mode == BatchMode.MERGED_QUADRATIC:
                log_item = WandBImagePlaceholder(np_make_grid(
                    images,
                    nrow=int(len(images) ** 0.5 + 0.49),
                    pad_value=255,
                    padding=5
                ))
            elif self.batch_mode == BatchMode.MERGED_ROW:
                log_item = WandBImagePlaceholder(np_make_grid(
                    images,
                    nrow=len(images),
                    pad_value=255,
                    padding=5
                ))
            elif self.batch_mode == BatchMode.MERGED_COLUMN:
                grid = np_make_grid(images, nrow=1, pad_value=255, padding=5)
                log_item = WandBImagePlaceholder(grid)
            elif self.batch_mode == BatchMode.SEPARATE:
                log_item = WandBImageListPlaceholder([image for image in images])
            else:
                raise ValueError(f"Unknown BatchMode: {self.batch_mode}")
            result[f"{self.log_name}_{name}_{visualization_name}"] = log_item

        return result


class WandBImageLogger(WandBMetricLogger):
    def __init__(
            self,
            *,
            log_name: str = "batch",
            epoch_interval: int = 100,
            log_final: bool = False,
            channels: slice = slice(0, 3),
            target_channels: slice = slice(None, None),
            prefix: Optional[str] = None,
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final, prefix=prefix)
        self.log_name = log_name
        self.channels = channels
        self.target_channels = target_channels
        warnings.warn("This logger is deprecated. Please use WandBBatchLogger instead!")

    def get_log_data(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        log_data = {
            f"{self.log_name}_state": [wandb.Image(img, mode=img.mode) for img in
                                       states_to_rgb_pillow(state_progression[-1], channels=self.channels)],
            f"{self.log_name}_seeds": [wandb.Image(img, mode=img.mode) for img in
                                       states_to_rgb_pillow(state_progression[0], channels=self.channels)],
            f"{self.log_name}_targets": [wandb.Image(img, mode=img.mode) for img in
                                         states_to_rgb_pillow(data_sample.targets, channels=self.target_channels)],
        }

        return log_data


@dataclass
class VideoParameter:
    image_slice: slice = field(default_factory=lambda: slice(None, None))
    fps: int = 24
    name: str = "video"
    batch_padding: int = 5

    def to_video_name(self, visualization_name: str) -> tuple[str, str]:
        """
        :param visualization_name:
        :return: tuple[log_name, caption]
        """
        caption_start = ""
        caption_end = ""
        caption_speedup = ""
        log_start = ""
        log_end = ""
        log_speedup = ""
        if self.image_slice.start is not None:
            caption_start = f" from {self.image_slice.start}"
            log_start = self.image_slice.start
        if self.image_slice.stop is not None:
            caption_end = f" until {self.image_slice.stop}"
            log_end = self.image_slice.stop
        if self.image_slice.step is not None and self.image_slice.step != 1:
            caption_speedup = f" (::{self.image_slice.step})"
            log_speedup = self.image_slice.step

        log_name = f"{self.name}_[{log_start}:{log_end}:{log_speedup}]_{visualization_name}"
        caption = f"{self.name}{caption_start}{caption_end}{caption_speedup} {visualization_name}"
        return log_name, caption


ProgressionParameter = slice | Iterable[int]


class WandBInferenceLogger(WandBMetricLogger):
    def __init__(
            self,
            seed_factory: NCA_SeedFactory,
            visualizer: StateVisualizer,
            *,
            epoch_interval: int = 1000,
            batch_size=1,
            video_parameters: Optional[VideoParameter | Iterable[VideoParameter]] = VideoParameter(slice(None, 1000)),
            progression_parameters: Optional[ProgressionParameter | Iterable[ProgressionParameter]] = None,
            static_image_parameters: Optional[int | Iterable[int]] = -1,
            max_steps: Optional[int] = None,
            log_final: bool = False,
            prefix: Optional[str] = None,
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final, prefix=prefix)
        # General
        self.batch_size = batch_size
        self.seed_factory = seed_factory
        self.visualizer = visualizer

        # Video parameter
        self.video_parameters, video_steps = self._initialize_video_parameters(video_parameters)

        # Progression images
        self.progression_parameters, prog_steps = self._initialize_progression_parameters(progression_parameters)

        # Static images
        self.static_image_parameters, image_steps = self._initialize_static_image_parameters(static_image_parameters)

        if len(self.video_parameters) + len(self.progression_parameters) + len(self.static_image_parameters) > 0:
            self._requires_images = True
        else:
            self._requires_images = False

        required_max_steps = max(video_steps, prog_steps, image_steps)
        if max_steps is not None:
            if max_steps < required_max_steps:
                self.logger.warn(f"max steps is set to {max_steps}, "
                                 f"but {required_max_steps} steps required according to parameters. "
                                 f"Steps greater than {max_steps} will be ignored!")
            self.steps = max_steps
        elif required_max_steps == 0:
            raise ValueError("max_steps is not given, but cannot be inferred from other parameters!")
        else:
            # Case: max_steps is None -> Take from required_max_steps
            self.steps = required_max_steps

    @staticmethod
    def _initialize_video_parameters(
            video_parameters: Optional[VideoParameter | Iterable[VideoParameter]]
    ) -> tuple[list[VideoParameter], int]:
        # Bring to correct shape
        if video_parameters is None:
            video_parameters = []
        elif isinstance(video_parameters, VideoParameter):
            video_parameters = [video_parameters]

        # Count max steps and validate VideoParameter
        max_steps = 0
        for video_parameter in video_parameters:
            if not isinstance(video_parameter, VideoParameter):
                raise TypeError(f"Expects type VideoParameter but got {type(video_parameter)}")

            video_stop = video_parameter.image_slice.stop
            if video_stop is not None:
                max_steps = max(video_stop, max_steps)

        return list(video_parameters), max_steps

    @staticmethod
    def _initialize_progression_parameters(
            progression_parameters: Optional[ProgressionParameter | Iterable[ProgressionParameter]]
    ) -> tuple[list[ProgressionParameter], int]:
        # Bring to correct shape
        if progression_parameters is None:
            progression_parameters = []
        elif isinstance(progression_parameters, slice):
            progression_parameters = [progression_parameters]
        if len(progression_parameters) > 0 and isinstance(progression_parameters[0], int):
            # Converting list of idx to list of list of idx, so we have a list of ProgressionParameter
            progression_parameters = [progression_parameters]

        # Count max steps and validate Progression Parameter
        max_steps = 0
        for progression_parameter in progression_parameters:
            if isinstance(progression_parameter, slice):
                progression_stop = progression_parameter.stop
                if progression_stop is not None:
                    max_steps = max(progression_stop, max_steps)
                continue

            # Expects an Iterable of ints here
            for idx in progression_parameter:
                if not isinstance(idx, int):
                    raise TypeError(f"Expects iterable of ints at this place for Progression Parameter, "
                                    f"but got {idx} ({progression_parameter})")
                max_steps = max(max_steps, idx)

        return list(progression_parameters), max_steps

    @staticmethod
    def _initialize_static_image_parameters(
            static_image_parameters: Optional[int | Iterable[int]]
    ) -> tuple[list[int], int]:
        # Bring to correct shape
        if static_image_parameters is None:
            static_image_parameters = []
        elif isinstance(static_image_parameters, int):
            static_image_parameters = [static_image_parameters]

        # Count max steps and validate static image parameters
        max_steps = 0
        for pos, idx in enumerate(static_image_parameters):
            if not isinstance(idx, int):
                raise TypeError(f"Expects ints at this place for static images, "
                                f"but got {idx} on position {pos} in {static_image_parameters}")

            max_steps = max(max_steps, idx)

        return static_image_parameters, max_steps

    def get_log_data(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        """ Do not use data_sample, state_progression or loss here, as these are not wanted in Inference Logger"""
        log_data = {}
        nca = self.trainer.nca
        seed_state = self.seed_factory(self.batch_size, device=nca.device)
        if self._requires_images:
            image_timelines, final_state = generate_video_array(
                nca,
                seed_state,
                steps=self.steps,
                visualizer=self.visualizer
            )
            log_data.update(self.create_videos(image_timelines))
            log_data.update(self.create_progression_images(image_timelines))
            log_data.update(self.create_static_images(image_timelines))
        else:
            final_state = run_inference(nca, seed_state, steps=self.steps)

        return log_data

    def create_videos(self, image_timelines: dict[str, np.ndarray]) -> dict[str, WandBVideoPlaceholder]:
        """
        :param image_timelines: dict[name, np.ndarray[T x B x W x H x 3]]
        :return: dict[str, wandb-video-placeholder]
        """
        log_data = {}
        for visualization_name, image_timeline in image_timelines.items():
            for video_parameter in self.video_parameters:
                log_name, caption = video_parameter.to_video_name(visualization_name)
                video = video_timeline_preparation(
                    image_timeline[video_parameter.image_slice],
                    padding=video_parameter.batch_padding
                )
                log_data[log_name] = WandBVideoPlaceholder(
                    video,
                    caption=caption,
                    format="gif",
                    fps=video_parameter.fps
                )

        return log_data

    def create_progression_images(self, image_timelines: dict[str, np.ndarray]) -> dict[str, WandBImageListPlaceholder]:
        """
        :param image_timelines: dict[name, np.ndarray[T x B x W x H x 3]]
        :return: dict[str, wandb-image-list-placeholder]. Creates per Batch one Progression
        """
        log_data = {}
        for visualization_name, image_timeline in image_timelines.items():
            for i, progression_parameter in enumerate(self.progression_parameters):
                # Get name for this progression_parameter
                progression_name = i
                if isinstance(progression_parameter, slice):
                    progression_name = str(progression_parameter)

                # Process progression
                progression_images = image_timeline[progression_parameter]
                # progression_images: np.ndarray[T' x B x W x H x 3]
                row, _ = as_quadratic_shape_as_possible(len(progression_images))
                merged_progression_images = [
                    np_make_grid(images, nrow=row, pad_value=255)
                    for images in progression_images.transpose((1, 0, 2, 3, 4))
                ]
                # merged_progression_images: list[np.ndarray[W x H x 3]] with length B

                log_data[f"progression_{progression_name}_{visualization_name}"] = WandBImageListPlaceholder(
                    merged_progression_images
                )

        return log_data

    def create_static_images(self, image_timelines: dict[str, np.ndarray]) -> dict[str, WandBImagePlaceholder]:
        """
        :param image_timelines: dict[name, np.ndarray[T x B x W x H x 3]]
        :return: dict[str, wandb-image-placeholder]
        """
        log_data = {}
        for visualization_name, image_timeline in image_timelines.items():
            nrow, _ = as_quadratic_shape_as_possible(image_timeline.shape[1])
            for static_image_idx in self.static_image_parameters:
                images = image_timeline[static_image_idx]
                # images: np.ndarray[B x W x H x 3]
                log_data[f"image_{static_image_idx}_{visualization_name}"] = WandBImagePlaceholder(
                    np_make_grid(
                        images, nrow=nrow, pad_value=255
                    )
                )

        return log_data


class WandBTrainingParameterLogger(WandBMetricLogger):
    def __init__(
            self,
            *,
            epoch_interval: int = 25,
            log_final: bool = True,
            prefix: Optional[str] = None,
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final, prefix=prefix)

    def get_log_data(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        log_data = {
            "learning_rate": self.trainer.optimizer.param_groups[-1]['lr']
        }
        return log_data


class WandBVideoLogger(WandBMetricLogger):
    def __init__(
            self,
            seed_factory: NCA_SeedFactory,
            *,
            fps: int = 24,
            steps: int = 300,
            log_name: str = "video",
            epoch_interval: int = 1000,
            log_final: bool = False,
            take_every: int = 1,
            rgb_channels: slice = slice(0, 3),
            has_alpha: bool = False,
            log_image=True,
            prefix: Optional[str] = None,
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final, prefix=prefix)
        warnings.warn("This logger is deprecated. Please use WandBInferenceLogger instead!")
        self.seed_factory = seed_factory

        self.fps = fps
        self.steps = steps
        self.take_every = take_every
        self.log_name = log_name
        self.log_image = log_image

        self.rgb_channels = rgb_channels
        self.has_alpha = has_alpha

    def init_training(self, trainer: NCA_Trainer, stopping_criterion: NCA_StoppingCriterion):
        super().init_training(trainer, stopping_criterion)
        wandb.watch(self.trainer.nca, log_freq=self.epoch_interval)

    def get_log_data(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        result = _legacy_generate_video_array(
            self.trainer.nca,
            self.seed_factory(),
            steps=self.steps,
            rgb_channels=self.rgb_channels
        )

        if result.shape[1] == 1:
            result = np.repeat(result, 3, axis=1)

        log_data = {
            self.log_name: wandb.Video(
                blend_alpha(result[::self.take_every]).astype("uint8") if self.has_alpha else result[::self.take_every],
                caption=f"Sample video ({self.steps} steps)",
                format="gif",
                fps=self.fps
            )
        }

        # Add final image
        if self.log_image:
            log_data[f"{self.log_name} (Image)"] = wandb.Image(
                result[-1].transpose(1, 2, 0),
                caption=f"Sample Image ({self.steps} steps)",
                mode="RGBA" if self.has_alpha else "RGB"
            )

        return log_data


class WandBValidationLogger(WandBMetricLogger):
    def __init__(
            self,
            dataset: NCA_Dataset,
            loss_function: NCA_LossFunction,
            *,
            epoch_interval: int = 1000,
            batch_size=1,
            log_final: bool = False,
            prefix: Optional[str] = None,
            eval_steps=(0, 300, 1000, 2000)
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final, prefix=prefix)
        self.dataset = dataset
        self.loss_function = loss_function
        self.steps = max(eval_steps)
        self.batch_size = batch_size
        self.eval_steps = list(eval_steps)
        self.batcher = SimpleBatcher(self.dataset, self.batch_size)

    def get_log_data(
            self,
            data_sample: DataSample,
            state_progression: torch.Tensor,
            loss: LossResult
    ) -> dict[str, Any]:
        log_data = {}
        with torch.no_grad():
            sample = self.batcher.next_batch()
            state = sample.states.detach()
            nca = self.trainer.nca
            kwargs = sample.get_kwargs(nca.required_kwargs())
            for step in range(self.steps):
                this_kwargs = {key: value[step] for key, value in kwargs.items()}

                state = nca(state, **this_kwargs)

                if (step + 1) in self.eval_steps:
                    loss = self.loss_function(state[None], sample.targets)
                    log_data[f"validation_{step + 1}"] = loss.total_loss(ignore_inf=False)

        return log_data



from typing import Any, Optional

import torch

from ncalib.nca.state_updater import NCA_StateUpdater
from ncalib.utils.torch_helper import update_part_of_state
from ncalib.utils.utils import slice_to_config


class SimpleStateUpdater(NCA_StateUpdater):
    def __init__(self, *, channels: slice = slice(None, None), scaling: float = 1., delta_channels=None):
        super().__init__()
        self.scaling = scaling
        self.channels = channels

        if delta_channels is None:
            delta_channels = channels

        self.delta_channels = delta_channels


    def forward(
            self,
            state: torch.Tensor,
            delta_state: torch.Tensor,
            **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        new_state = state[:, self.channels] + self.scaling * delta_state[:, self.delta_channels]
        state = update_part_of_state(state, new_state, self.channels)

        return state, delta_state

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["scaling"] = self.scaling
        cfg["channels"] = slice_to_config(self.channels)
        cfg["delta_channels"] = slice_to_config(self.delta_channels)
        return cfg


class FixedStatesUpdater(NCA_StateUpdater):
    def __init__(self, fixed_channels: int | slice):
        super().__init__()
        if isinstance(fixed_channels, int):
            fixed_channels = slice(fixed_channels, fixed_channels + 1)
        self.fixed_channels = fixed_channels

    def forward(
            self,
            state: torch.Tensor,
            delta_state: torch.Tensor,
            *,
            original_state: Optional[torch.Tensor] = None,
            **kwargs
    ) -> tuple[torch.Tensor, torch.Tensor]:
        state[:, self.fixed_channels] = original_state[:, self.fixed_channels]
        return state, delta_state

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["fixed_channels"] = slice_to_config(self.fixed_channels)
        return cfg

    def required_kwargs(self) -> set[str]:
        res = super().required_kwargs()
        res.add("original_state")
        return res





from typing import Any

import torch
import torch.nn.functional as F
from torch import nn

from ncalib.nca.models import NCA_Model


class TinyNCAModel(NCA_Model):
    def __init__(self, input_size: int, output_size: int):
        super().__init__(input_size, output_size)

        self.conv1 = nn.Conv2d(input_size, 8, 1)
        self.conv2 = nn.Conv2d(input_size, 8, 1)

    def forward(self, perception: list[torch.Tensor], **kwargs) -> torch.Tensor:
        perception = torch.cat(perception, 1)
        return self.conv2(F.relu(self.conv1(perception)))

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        return cfg



import abc

import numpy as np
import torch

from ncalib.utils.utils import NCA_Base


class StateVisualizer(NCA_Base, abc.ABC):
    def __init__(self):
        super().__init__()

    @abc.abstractmethod
    def __call__(self, state: torch.Tensor) -> dict[str, np.ndarray]:
        """
        :param state: Tensor[B x C x W x H]
        :return: dictionary of {name: np.ndarray[B x W x H x 1/3/4]}
        """
        pass

    def __add__(self, other: "StateVisualizer") -> "StateVisualizer":
        # Prepare each visualizer so that each is an iterable of simple visualizations
        if isinstance(self, CombinedVisualizer):
            vis1 = self.visualizer
        else:
            vis1 = [self]

        if isinstance(other, CombinedVisualizer):
            vis2 = other.visualizer
        else:
            vis2 = [other]

        # Combine Visualizer
        return CombinedVisualizer(*vis1, *vis2)


class CombinedVisualizer(StateVisualizer):
    def __init__(
            self,
            *visualizer: StateVisualizer,
    ):
        super().__init__()
        self.visualizer: tuple[StateVisualizer] = tuple(visualizer)

    def __call__(self, state: torch.Tensor) -> dict[str, np.ndarray]:
        res = {}
        for visualizer in self.visualizer:
            res.update(visualizer(state))
        return res


class BatchedStateVisualizer(StateVisualizer, abc.ABC):
    def __init__(
            self,
            name: str,
            batch_slice=slice(None, None, None),
    ):
        super().__init__()
        self.name = name
        self.batch_slice: slice = batch_slice

    def __call__(self, state: torch.Tensor) -> dict[str, np.ndarray]:
        """
        :param state: Tensor[B x C x W x H]
        :return: dictionary of {name: np.ndarray[B x W x H x 1/3/4]} with ints 0.255 for Gray, BGR, BGRA
        """
        batch = state[self.batch_slice].detach()
        batch_size = batch.shape[0]

        visualization = self._batch_to_visualization(batch)
        # visualization is dict with key=name and value either:
        #   -> torch.Tensor[B x 1/3/4 x W' x H'] with floats 0..1 for Gray, RGB, RGBA
        #   -> np.ndarray[B x W' x H' x 1/3/4] with ints 0..255 for Gray, BGR, BGRA

        result = {}
        for key, image in visualization.items():
            if isinstance(image, torch.Tensor):
                image = (torch.clip(image, 0, 1) * 255).cpu().numpy().astype("uint8")  # 0..1 -> 0.255
                image = image.transpose((0, 2, 3, 1))  # [B x C x W' x H'] -> [B x W' x H' x C]

            if not isinstance(image, np.ndarray):
                raise TypeError(f"Expects image to be np.ndarray (it is {type(image)})")

            if not image.shape[3] in [1, 3, 4]:
                raise ValueError(f"Expects number of color channels to be 1,3 or 4 (image has shape of {image.shape})")

            result[key] = image
        return result

    @abc.abstractmethod
    def _batch_to_visualization(self, batch: torch.Tensor) -> dict[str, torch.Tensor | np.ndarray]:
        """
        :param batch: batched Tensor [B x C x W x H]
        :return: dict[
            name ->
                torch.Tensor[B x 1/3/4 x W x H] with float ranges 0..1 in Gray, RGB, RGBA
                np.ndarray[B x W x H x 1/3/4] with int ranges 0..255 in Gray, BGR, BGRA
            ]
        """
        pass



import abc
from typing import Optional

import torch

from ncalib.trainer.dataset import DataSample
from ncalib.trainer.loss_function import LossResult
from ncalib.utils.utils import NCA_Base


class NCA_StoppingCriterion(NCA_Base, abc.ABC):

    def __init__(self, message: str = "", is_fatal: bool = False):
        super().__init__()
        self.message = message
        self.is_fatal = is_fatal
        self.last_evaluation = False

    def __call__(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        """
        :param trainer:
        :param sample: DataSample with at least .states[B x C x W x H], .targets[B x C' x W x H] and .num_steps[B]
        :param state_progression: Tensor[T x B x C x W x H]
        :param loss: LossResult
        :return: True if it should stop, False otherwise
        """
        self.last_evaluation = self.evaluate(trainer, sample, state_progression, loss)
        return self.last_evaluation

    def __neg__(self) -> "ChainedNot":
        return ChainedNot(self)

    def __or__(self, other: "NCA_StoppingCriterion") -> "ChainedOr":
        assert isinstance(other, NCA_StoppingCriterion)

        return ChainedOr(self, other)

    def __and__(self, other: "NCA_StoppingCriterion") -> "ChainedAnd":
        assert isinstance(other, NCA_StoppingCriterion)

        return ChainedAnd(self, other)

    def __len__(self):
        return None

    @abc.abstractmethod
    def evaluate(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        """
        :param trainer:
        :param sample: DataSample with at least .states[B x C x W x H], .targets[B x C' x W x H] and .num_steps[B]
        :param state_progression: Tensor[T x B x C x W x H]
        :param loss: LossResult
        :return: True if it should stop, False otherwise
        """
        pass


class ChainedOr(NCA_StoppingCriterion):
    def __init__(self, *criteria: NCA_StoppingCriterion):
        super().__init__()
        self.criteria = criteria

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{self.criteria}"

    def __add__(self, other) -> "ChainedOr":
        assert isinstance(other, NCA_StoppingCriterion)

        return ChainedOr(*self.criteria, other)

    def __len__(self) -> Optional[int]:
        values = list(filter(lambda s: s is not None, map(lambda s: s.__len__(), self.criteria)))
        if len(values) == 0:
            return None
        return min(values)

    def evaluate(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        for criterion in self.criteria:
            if criterion(trainer, sample, state_progression, loss):
                self.is_fatal = criterion.is_fatal
                self.message = criterion.message
                return True

        self.is_fatal = False
        self.message = ""
        return False


class ChainedAnd(NCA_StoppingCriterion):
    def __init__(self, *criteria: NCA_StoppingCriterion):
        super().__init__()
        self.criteria = criteria
        self.is_fatal = any(criterion.is_fatal for criterion in criteria)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}{self.criteria}"

    def __len__(self) -> Optional[int]:
        values = list(filter(lambda s: s is not None, map(lambda s: s.__len__(), self.criteria)))
        if len(values) == 0:
            return None
        return max(values)

    def __mul__(self, other) -> "ChainedAnd":
        assert isinstance(other, NCA_StoppingCriterion)

        return ChainedAnd(*self.criteria, other)

    def evaluate(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        message = ""
        for criterion in self.criteria:
            if not criterion(trainer, sample, state_progression, loss):
                return False

            message += f"{criterion.message}\n"

        self.message = message
        return True


class ChainedNot(NCA_StoppingCriterion):
    def __init__(self, criterion: NCA_StoppingCriterion):
        super().__init__()
        self.criterion = criterion
        self.message = self.criterion.message
        self.is_fatal = self.criterion.is_fatal

    def __neg__(self) -> NCA_StoppingCriterion:
        return self.criterion

    def evaluate(self, trainer: "NCA_Trainer", sample: DataSample, state_progression: torch.Tensor,
                 loss: LossResult) -> bool:
        return not self.criterion(trainer, sample, state_progression, loss)



import json
import time
import warnings
from pathlib import Path

import torch

from ncalib.trainer import NCA_Logger
from ncalib.trainer.dataset import DataSample
from ncalib.trainer.loss_function import LossResult


class ModelLogger(NCA_Logger):
    def __init__(
            self,
            log_dir: str | Path,
            model_name: str = None,
            *,
            leading_zeros: int = 5,
            epoch_interval: int = 1000,
            log_final: bool = False
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final)
        self.log_dir = Path(log_dir)

        if model_name is None:
            model_name = f"model-{time.time():.0f}"

        self.model_name = model_name
        self.leading_zeros = leading_zeros
        self._latest_model_path = None

    @property
    def latest_model_path(self) -> Path:
        if self._latest_model_path is None:
            raise FileNotFoundError("No model saved yet!")
        return self._latest_model_path

    def init_training(self, trainer: "NCA_Trainer", stopping_criterion: "NCA_StoppingCriterion"):
        super().init_training(trainer, stopping_criterion)
        self.log_dir.mkdir(exist_ok=True, parents=True)

    def _log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        full_config = self.trainer.nca.as_full_config_dict()
        log_file = self.log_dir / f"{self.model_name}.{self.trainer.epoch:0{self.leading_zeros}d}.nca"

        self.logger.debug(f"Saving model for epoch {self.trainer.epoch} at \"{log_file}\"")
        torch.save(full_config, log_file)
        self._latest_model_path = log_file


class LossJSONLinesLogger(NCA_Logger):
    def __init__(
            self,
            filename: Path | str,
            epoch_interval: int = 10,
            log_final: bool = True
    ):
        super().__init__(epoch_interval=epoch_interval, log_final=log_final)
        self.filename = Path(filename)
        self.opened_file = None

        self._i = 0

    def init_training(self, trainer: "NCA_Trainer", stopping_criterion: "NCA_StoppingCriterion"):
        super().init_training(trainer, stopping_criterion)
        self.filename.parent.mkdir(exist_ok=True, parents=True)
        self.opened_file = self.filename.open("w")
        self._i = 0

    def _log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        detailed_loss = loss.detailed_loss(include_total=True)
        log_data = {
            "epoch": self.trainer.epoch
        }
        log_data.update({f"loss_{name}": loss for name, (loss, weight) in detailed_loss.items()})
        log_data.update({f"lossweighted_{name}": loss * weight for name, (loss, weight) in detailed_loss.items() if
                         name != "total"})
        if self.opened_file is not None:
            self.opened_file.write(json.dumps(log_data) + "\n")
            self._i += 1
            if self._i % 20 == 0:
                self.opened_file.flush()
        else:
            warnings.warn(f"File {self.filename} not opened! Skip logging...")

    def close(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        if self.opened_file is not None:
            self.opened_file.close()



import torch
from jaxtyping import Float

from ncalib import NCA_Perception


class SteerablePerception(NCA_Perception):
    def __init__(
            self,
            perception: NCA_Perception,
            channel_id: int,
            *,
            channel_slice:slice=slice(None,None,None),
    ):
        super().__init__(channel_slice=channel_slice)
        self.channel_id = channel_id
        self.perception = perception

    def forward(self, state: Float[torch.Tensor, "B C W H"], **kwargs) -> list[Float[torch.Tensor, "B C W H"]]:
        # TODO: Can we just remove this or implement in every perception?
        channel_mask = torch.arange(state.shape[1]) != self.channel_id

        # Calculate original perception
        perceptions = self.perception(state[:, channel_mask])
        if len(perceptions) != 2:
            raise RuntimeError("Internal perception needs to return a list of 2 Perception matrices")

        p1, p2 = perceptions
        angle = state[:, self.channel_id]
        cos_angle, sin_angle = torch.cos(angle)[:, None], torch.sin(angle)[:, None]
        return [p1 * cos_angle + p2 * sin_angle, p2 * cos_angle - p1 * sin_angle]

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["perception"] = self.perception.to_config_dict()
        cfg["channel_id"] = self.channel_id
        return cfg



import torch

from ncalib.trainer.batcher.pool import PoolEvictionPolicy
from ncalib.trainer.dataset import DataSample


class RandomEvictionPolicy(PoolEvictionPolicy):
    def score_batch(self, data_sample: DataSample, losses: torch.Tensor) -> torch.Tensor:
        return torch.zeros_like(losses)



import math
from pathlib import Path
from typing import Callable, Any, Optional, Iterable

import torch
import torch.nn as nn
from torch.utils.data import Dataset, IterableDataset, DataLoader
from torchvision.datasets import VisionDataset
from torchvision.datasets.folder import default_loader


def weight_init(m):
    """Custom weight init for Conv2D and Linear layers."""
    if isinstance(m, nn.Linear):
        nn.init.orthogonal_(m.weight.data)
        if hasattr(m.bias, "data"):
            m.bias.data.fill_(0.0)
    elif isinstance(m, nn.Conv2d) or isinstance(m, nn.ConvTranspose2d):
        gain = nn.init.calculate_gain("relu")
        nn.init.orthogonal_(m.weight.data, gain * 1e-2)
        if hasattr(m.bias, "data"):
            m.bias.data.fill_(0.0)


def reset_conv2d(layer):
    n = layer.in_channels
    for k in layer.kernel_size:
        n *= k
    stdv = 1. / math.sqrt(n)
    layer.weight.data.uniform_(-stdv, stdv)
    if layer.bias is not None:
        layer.bias.data.uniform_(-stdv, stdv)


def cleanup_state_dict(state_dict) -> dict[str, torch.Tensor]:
    """
    Removes _orig_mod. which is introduced by torch.compile()
    :param state_dict:
    :return:
    """
    new_state_dict = dict()
    for key in state_dict:
        new_state_dict[key.replace("_orig_mod.", "")] = state_dict[key]

    return new_state_dict


def update_part_of_state(state: torch.Tensor, new_state: torch.Tensor, channels: slice) -> torch.Tensor:
    """
    Replaces a part of a state in a Torch-friendly manner so that gradients can be back propagated.
    :param state: original state that should be modified Tensor[B x C x W x H]
    :param new_state: new state only for the range that should be replaced [B x C' x W x H]
    :param channels: slice with the respective channels. Covers C' channels
    :return: Tensor[B x C x W x H]
    """
    device = state.device
    B, C, W, H = state.shape

    # Avoid replacements and circumvent by multiplying by 0 and adding new state
    with torch.no_grad():
        ones_output = torch.zeros((1, C, 1, 1), device=device)
        ones_output[:, channels] = 1
        zeros_output = 1 - ones_output

        expanded_new_state = torch.zeros_like(state, device=device)

    expanded_new_state[:, channels] = new_state

    return state * zeros_output + expanded_new_state * ones_output


class SimpleImageDataset(VisionDataset):
    def __init__(
            self,
            *paths: Path | str,
            transform: Optional[Callable] = None,
            loader: Callable[[str], Any] = default_loader,
            use_cache: bool = True
    ):
        super().__init__(Path().root, transform=transform)
        self.loader = loader
        self.paths = SimpleImageDataset.parse_paths(paths)
        self.use_cache = use_cache
        self._cache = {}

    def __getitem__(self, index) -> torch.Tensor:
        if index in self._cache and self.use_cache:
            return self._cache[index]

        path = self.paths[index]
        sample = self.loader(str(path))
        if self.transform is not None:
            sample = self.transform(sample)
        if self.use_cache:
            self._cache[index] = sample
        return sample

    def __len__(self):
        return len(self.paths)

    @classmethod
    def parse_paths(cls, paths: Iterable[Path | str]) -> list[Path]:
        result = []
        for path in paths:
            path = Path(path)
            if path.is_dir():
                result += [p for p in path.iterdir() if p.is_file()]
            else:
                result.append(path)

        return result


class EmitTargetsVisionDataset(VisionDataset):
    def __init__(self, dataset: VisionDataset):
        super().__init__("")
        self.dataset = dataset

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, item):
        image, target = self.dataset.__getitem__(item)
        return image

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.dataset.__repr__()})"


class EndlessIterableDataset(IterableDataset):
    def __init__(self, dataset: Dataset, *, generator: torch.Generator = None, shuffle=False):
        self.dataset = dataset
        if generator is None:
            self.rng = torch.Generator()
        else:
            self.rng = generator
        self.shuffle = shuffle

    def __iter__(self):
        dataloader = DataLoader(self.dataset, batch_size=None, shuffle=self.shuffle, generator=self.rng)
        while True:
            for item in dataloader:
                yield item



from typing import Callable

import torch
from torch.utils.data import DataLoader

from ncalib.trainer.batcher import NCA_Batcher
from ncalib.trainer.dataset import NCA_Dataset, DataSample
from ncalib.utils.defaults import DEFAULT_DEVICE


class SimpleBatcher(NCA_Batcher):
    def __init__(
            self,
            dataset: NCA_Dataset,
            batch_size: int,
            num_workers: int = 0,
            collate_fn: Callable = DataSample.merge_batch,
            device: torch.device | str = DEFAULT_DEVICE
    ):
        super().__init__(dataset, batch_size, num_workers=num_workers, collate_fn=collate_fn, device=device)
        self.dataloader = iter(DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=False,  # Shuffling is done in Dataset
            num_workers=num_workers,
            drop_last=True,
            collate_fn=collate_fn
        ))

    def next_batch(self) -> DataSample:
        return next(self.dataloader).to(self.device)



import abc
from typing import Iterable, Any, Optional

import torch
from jaxtyping import Float

from ncalib.utils.utils import NCA_Base


class LossResult:
    def __init__(
            self,
            loss_tensor: torch.Tensor,
            *,
            names: Iterable[str] = ("Loss",),
            weights: Optional[torch.Tensor | Iterable[float] | float] = None
    ):
        """
        :param loss_tensor: torch.Tensor[B x L] or torch.Tensor[B], assuming a single loss
        :param names: list[str] of length L
        """
        if len(loss_tensor.shape) == 1:
            loss_tensor = loss_tensor[:, None]
            # Should now be shape [B x 1]

        if len(loss_tensor.shape) != 2:
            raise ValueError(f"Loss-Tensor has invalid shape. Should be [B x L], but is {loss_tensor.shape}")

        self.loss_tensor = loss_tensor  # Shape [B x L]
        self.names = tuple(names)
        if weights is None:
            weights = torch.ones(len(self.names))
        elif isinstance(weights, (int, float)):
            weights = (weights,)
        self.weights: torch.Tensor = torch.as_tensor(weights)  # Shape[L]

        if len(self.names) != self.loss_tensor.shape[1]:
            raise ValueError(f"Invalid lengths for loss: names has {len(self.names)} values, "
                             f"but loss-tensor has shape {loss_tensor.shape}")
        if len(self.weights) != self.loss_tensor.shape[1]:
            raise ValueError(f"Invalid lengths for loss: weights has {len(self.weights)} values, "
                             f"but loss-tensor has shape {loss_tensor.shape}")

    def __add__(self, other: "LossResult") -> "LossResult":
        if not isinstance(other, LossResult):
            raise TypeError(f"Expects other to also be a {self.__class__.__name__} when adding (it is {type(other)}).")

        all_names = list(self.names)
        for original_name in other.names:
            name = original_name
            i = 0
            while name in all_names:
                i += 1
                name = f"{original_name}_{i}"

            all_names.append(name)

        loss_tensor = torch.cat([self.loss_tensor, other.loss_tensor], dim=1)
        return LossResult(loss_tensor, names=all_names, weights=torch.cat([self.weights, other.weights]))

    def __mul__(self, value: float) -> "LossResult":
        self.loss_tensor *= value
        return self

    def __truediv__(self, other: float) -> "LossResult":
        return self.__mul__(1 / other)

    def __rmul__(self, value: float) -> "LossResult":
        return self.__mul__(value)

    def __repr__(self):
        return f"{self.__class__.__name__}(shape={self.loss_tensor.shape}, total_loss={self.total_loss()})"

    def __getitem__(self, item) -> torch.Tensor:
        idx = self.names.index(item)
        return self.loss_tensor[:, idx] * self.weights[idx]

    @classmethod
    def from_dict(
            cls,
            data: dict[str, Float[torch.Tensor, "B"]],
            weights: Optional[Iterable[float] | float] = None
    ) -> "LossResult":
        """
        :param data: dict[str, Tensor[B]]
        :return: LossResult
        """
        stacked_data = torch.stack(list(data.values()), dim=1)
        return LossResult(stacked_data, names=data.keys(), weights=weights)

    def total_loss(self, *, ignore_inf=False) -> Float[torch.Tensor, "1"]:
        """
        :return: Tensor(1)
        """
        batchwise_loss = self.batchwise_loss()
        if ignore_inf:
            return torch.mean(batchwise_loss[torch.isfinite(batchwise_loss)])
        return torch.mean(batchwise_loss)

    def batchwise_loss(self) -> Float[torch.Tensor, "B"]:
        """
        :return: Tensor[B]
        """
        return torch.sum(self.loss_tensor * self.weights[None].to(self.loss_tensor.device), dim=1)

    def detailed_loss(self, include_total=False) -> dict[str, tuple[float, float]]:
        """
        :return: dict[name, [loss_value, weight]]
        """
        detailed_loss = {
            name: (torch.mean(batch_loss).item(), weight.item())
            for name, weight, batch_loss in
            zip(self.names, self.weights, self.loss_tensor.detach().cpu().transpose(0, 1))
        }
        if include_total and len(detailed_loss) > 1:
            detailed_loss["total"] = self.total_loss().detach().cpu().item(), 1
        return detailed_loss

    def as_unweighted(self):
        return LossResult(self.loss_tensor, names=self.names, weights=None)


class NCA_LossFunction(NCA_Base, abc.ABC):
    def __init__(self, *, weight: float = 1.):
        super().__init__()
        self.weight = weight

    @abc.abstractmethod
    def __call__(
            self,
            state_progression: Float[torch.Tensor, "N B C W H"],
            target: Float[torch.Tensor, "B ..."],
            **kwargs
    ) -> LossResult:
        pass

    def __add__(self, other: "NCA_LossFunction") -> "CombinedLossFunction":
        # Prepare each loss function so that each is an iterable of simple loss functions
        if isinstance(self, CombinedLossFunction):
            losses1 = tuple(self.loss_functions)
        else:
            losses1 = (self,)

        if isinstance(other, CombinedLossFunction):
            losses2 = tuple(other.loss_functions)
        else:
            losses2 = (other,)

        # Combine Loss function
        return CombinedLossFunction(losses1 + losses2)

    def __mul__(self, value: float) -> "NCA_LossFunction":
        if not isinstance(value, (float, int)):
            raise TypeError(f"value has to be a number (it is {type(value)})")

        self.weight *= value
        return self

    def __truediv__(self, other: float) -> "NCA_LossFunction":
        return self.__mul__(1 / other)

    def __rmul__(self, value: float) -> "NCA_LossFunction":
        return self.__mul__(value)

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["weight"] = self.weight
        return cfg


class NamedLossFunction(NCA_LossFunction, abc.ABC):
    def __init__(self, loss_name, *, weight=1):
        super().__init__(weight=weight)
        self.loss_name = loss_name

    def __call__(
            self,
            state_progression: Float[torch.Tensor, "N B C W H"],
            target: Float[torch.Tensor, "B ..."],
            **kwargs
    ) -> LossResult:
        return LossResult.from_dict({
            self.loss_name: self._calculate_loss(state_progression, target, **kwargs)
        }, weights=self.weight)

    @abc.abstractmethod
    def _calculate_loss(
            self,
            state_progression: Float[torch.Tensor, "N B C W H"],
            target: Float[torch.Tensor, "B ..."],
            **kwargs
    ) -> Float[torch.Tensor, "B"]:
        pass

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["loss_name"] = self.loss_name
        return cfg


class CombinedLossFunction(NCA_LossFunction):
    def __init__(self, loss_functions):
        super().__init__()
        self.loss_functions = loss_functions

    def __call__(
            self,
            state_progression: Float[torch.Tensor, "N B C W H"],
            target: Float[torch.Tensor, "B ..."],
            **kwargs
    ) -> LossResult:
        result = self.loss_functions[0](state_progression, target, **kwargs)
        for loss_function in self.loss_functions[1:]:
            result += loss_function(state_progression, target, **kwargs)

        return result

    def __mul__(self, value: float):
        for loss_function in self.loss_functions:
            loss_function *= value
        return self

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["loss_functions"] = [loss_function.to_config_dict() for loss_function in self.loss_functions]
        return cfg



import sys
from typing import Optional

import torch
from tqdm.auto import tqdm

from ncalib.trainer import NCA_StoppingCriterion, NCA_Trainer
from ncalib.trainer.dataset import DataSample
from ncalib.trainer.logger import NCA_Logger
from ncalib.trainer.loss_function import LossResult


class TQDMLogger(NCA_Logger):
    def __init__(self, *, file=sys.stdout):
        super().__init__(epoch_interval=1)
        self.file = file

        self._pbar: Optional[tqdm] = None

    @property
    def progress_bar(self) -> tqdm:
        if self._pbar is None:
            self.not_initialized_yet()
        return self._pbar

    def init_training(self, trainer: NCA_Trainer, stopping_criterion: NCA_StoppingCriterion):
        super().init_training(trainer, stopping_criterion)
        total = len(stopping_criterion)
        self._pbar = tqdm(total=total, file=self.file)

    def _log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        self.progress_bar.update(1)
        self.progress_bar.set_postfix({
            name: f"{loss:.10f}"
            for name, (loss, weight)
            in loss.detailed_loss(include_total=True).items()
        })

    def close(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        super().close(data_sample, state_progression, loss)
        self.progress_bar.close()



import cv2
import torch
from tqdm import tqdm

from ncalib import NCA
from ncalib.seed_factory import NCA_SeedFactory
from ncalib.visualization.gui.window_manager import GUIWindow


class FrameBuffer(list):
    def __init__(self, size: int):
        super().__init__()
        self.size = size

    def append(self, frame: torch.Tensor):
        super().append(frame)
        self._cleanup()

    def _cleanup(self):
        while len(self) > self.size:
            self.pop(0)

    def get(self) -> torch.Tensor:
        return self.pop(-1)


class NCA_GUI:
    def __init__(
            self,
            nca: NCA,
            windows: GUIWindow,
            seed_factory: NCA_SeedFactory,
            *,
            batch_size: int = 1,
            initial_delay: int = 10,
            take_every_k_frames: int = 1,
            buffersize=100,
    ):
        self.nca = nca
        self.windows = windows
        self.seed_factory = seed_factory
        self.batch_size = batch_size

        self.delay = initial_delay
        self.take_every_k_frames = take_every_k_frames
        self.buffer = FrameBuffer(buffersize)

        # GUI features
        self.state = None
        self.paused = None
        self.pbar = None
        self.i = None

        self._required_kwargs = nca.required_kwargs()

    def run(self, n=None):
        device = self.nca.device
        with torch.no_grad():
            self.state = self.seed_factory(self.batch_size, device=device)

        self.windows.init_state(self.state)

        self.nca.eval()
        self.paused = -1
        self.i = 0
        self.buffer.clear()
        self.pbar = tqdm(desc="GUI", unit=" generations", unit_scale=True, total=n)
        while True:
            if self.paused != 0:
                self.pbar.update(self.take_every_k_frames)
                self.buffer.append(self.state)

                with torch.no_grad():
                    for k in range(self.take_every_k_frames):
                        kwargs = {}
                        self.i += 1
                        if "t" in self._required_kwargs and self.paused != 0:
                            if n is None:
                                raise ValueError("'t' is an required kwarg but no 'n' is given!")

                            if self.i > n:
                                break
                            kwargs["t"] = torch.as_tensor([1 - (self.i / n)] * self.batch_size)

                        self.state = self.nca(self.state, **kwargs)
                        self.windows.update_state(self.state, step=self.i)
                        self.paused -= 1
                        if self.paused == 0:
                            break
                    if n is not None and self.i > n:
                        self.paused = 0
                    self.state = self.windows.modify_state(self.state)

            self.windows.update_windows()
            key = cv2.waitKey(self.delay)
            try:
                self.handle_key(key)
                self.windows.handle_key(key)
            except StopIteration:
                break

        cv2.destroyAllWindows()
        self.pbar.close()

    def handle_key(self, key: int):
        if key != -1:
            self.pbar.set_postfix({f"Key pressed": key})

        if key == ord("q"):
            self.pbar.set_postfix({"cmd": f"Quit"})
            raise StopIteration()
        elif key == ord("r"):
            self.state = self.seed_factory(self.batch_size, device=self.state.device)
            self.windows.init_state(self.state)
            self.buffer.append(self.state)
            self.pbar.set_postfix({"cmd": f"Reset after {self.pbar.n} generations"})
            self.pbar.reset()
            self.i = 0
        elif key == ord("p") or key == ord(" "):
            if self.paused != 0:
                self.paused = 0
                self.pbar.set_postfix({"cmd": f"Paused"})
            else:
                self.paused = -1
                self.pbar.set_postfix({"cmd": f"Unpaused"})
        elif key == ord("+"):
            self.delay = max(int(self.delay / 1.5), 1)
            self.pbar.set_postfix({"cmd": f"Delay set to {self.delay}"})
        elif key == ord("-"):
            self.delay = max(int(self.delay * 1.5), 2)
            self.pbar.set_postfix({"cmd": f"Delay set to {self.delay}"})
        elif key == ord(">"):
            self.take_every_k_frames += 1
            self.pbar.set_postfix({"cmd": f"Take every {self.take_every_k_frames} frames"})
        elif key == ord("<"):
            self.take_every_k_frames = max(1, self.take_every_k_frames - 1)
            self.pbar.set_postfix({"cmd": f"Take every {self.take_every_k_frames} frames"})
        elif key == 81:  # Left Arrow
            if len(self.buffer) == 0:
                self.pbar.set_postfix({"cmd": f"Buffer empty!"})
                return
            self.state = self.buffer.get()
            self.i -= 1
            self.pbar.update(-1)
            self.pbar.set_postfix({"cmd": f"Taking next buffered image (new length={len(self.buffer)})"})
        elif key == 83:  # Right Arrow
            self.paused = 1



import torch
import torchvision.datasets

from ncalib.seed_factory import NCA_SeedFactory
from ncalib.utils.defaults import DEFAULT_DEVICE
from ncalib.utils.utils import endless_generator, slice_to_config


class DatasetSeedFactory(NCA_SeedFactory):
    def __init__(
            self,
            dataset: torchvision.datasets.VisionDataset,
            *,
            image_channels: slice,
            width=None,
            channels=16,
            height=None,
            num_workers=2
    ):
        super().__init__(width=width, channels=channels, height=height)
        self.dataset = dataset
        self.dataloader = endless_generator(torch.utils.data.DataLoader(
            self.dataset,
            batch_size=1,
            shuffle=True,
            num_workers=num_workers,
        ))

        self.image_channels = image_channels

    def __call__(self, n: int = 1, *, device=DEFAULT_DEVICE) -> torch.Tensor:
        images = []
        dimensions = torch.zeros((n, 3), dtype=int)
        for i in range(n):
            image, label = next(self.dataloader)
            image = image[0]
            images.append(image)
            dimensions[i] = torch.as_tensor(image.shape)

        max_dimension = torch.max(dimensions, dim=0).values
        if self.width is not None:
            if self.width < max_dimension[0]:
                raise OverflowError(f"Fixed width is too small for images! (needs at least w,h of {max_dimension})")
            max_w = self.width
        else:
            max_w = max_dimension[1]

        if self.height is not None:
            if self.height < max_dimension[1]:
                raise OverflowError(f"Fixed height is too small for images! (needs at least w,h of {max_dimension})")
            max_h = self.height
        else:
            max_h = max_dimension[2]

        result = torch.zeros((n, self.channels, max_w, max_h))
        for i, (image, (c, w, h)) in enumerate(zip(images, dimensions)):
            pad_x = int((max_w - w) // 2)
            pad_y = int((max_h - h) // 2)
            result[i, self.image_channels, pad_x:pad_x + w, pad_y:pad_y + h] = image

        return result.to(device)

    def to_config_dict(self):
        cfg = super().to_config_dict()
        cfg["image_channels"] = slice_to_config(self.image_channels)
        return cfg



import abc
from typing import Optional, Iterable, Union, Any

import torch

from ncalib.trainer.dataset import DataSample
from ncalib.trainer.loss_function import LossResult
from ncalib.utils.utils import NCA_Base


class NCA_Logger(NCA_Base, abc.ABC):
    def __init__(self, *, epoch_interval: int = 1000, log_final: bool = False):
        super().__init__()
        self.epoch_interval = epoch_interval
        self.log_final = log_final
        self._latest_logging_epoch = None

        # Training specific stuff
        self._trainer: Optional["NCA_Trainer"] = None
        self._stopping_criterion: Optional["NCA_StoppingCriterion"] = None

    def __add__(self, other: "NCA_Logger") -> "NCA_Logger":
        # Prepare each logger so that each is an iterable of simple logger
        if isinstance(self, CombinedLogger):
            loggers1 = self.loggers
        else:
            loggers1 = [self]

        if isinstance(other, CombinedLogger):
            loggers2 = other.loggers
        else:
            loggers2 = [other]

        # Combine Logger
        return CombinedLogger(*loggers1, *loggers2)

    def not_initialized_yet(self):
        raise RuntimeError(f"{self.__class__.__name__} not initialized yet. Please run `init_training()`")

    @property
    def trainer(self) -> "NCA_Trainer":
        if self._trainer is None:
            self.not_initialized_yet()
        return self._trainer

    @property
    def stopping_criterion(self) -> "NCA_StoppingCriterion":
        if self._stopping_criterion is None:
            self.not_initialized_yet()
        return self._stopping_criterion

    def init_training(self, trainer: "NCA_Trainer", stopping_criterion: "NCA_StoppingCriterion"):
        if self._trainer is not None or self._stopping_criterion is not None:
            raise RuntimeError("Logger already initialized. "
                               "Only use 1 instance of logger per training or it might yield to complications")
        self._stopping_criterion = stopping_criterion
        self._trainer = trainer

    def log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        """
        :param data_sample: DataSample with at least .states, .targets, .num_steps
        :param state_progression: Tensor[T x B x C x W x H]
        :param loss: LossResult
        :return:
        """
        if self.trainer.epoch % self.epoch_interval == 0:
            self._latest_logging_epoch = self.trainer.epoch
            return self._log(data_sample, state_progression, loss)

    @abc.abstractmethod
    def _log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        """
        :param data_sample: DataSample with at least .states, .targets, .num_steps
        :param state_progression: Tensor[T x B x C x W x H]
        :param loss: LossResult
        :return:
        """
        pass

    def close(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        if self.log_final and self.trainer.epoch != self._latest_logging_epoch:
            return self._log(data_sample, state_progression, loss)

    @classmethod
    def as_loggers_list(cls, loggers: Union["NCA_Logger", Iterable["NCA_Logger"], None]) -> list["NCA_Logger"]:
        if loggers is None:
            return []
        elif isinstance(loggers, NCA_Logger):
            return [loggers]

        for logger in loggers:
            if not isinstance(logger, cls):
                raise TypeError(f"{logger} is not a {cls.__name__}")

        return list(loggers)


class NoOPLogger(NCA_Logger):
    def _log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        pass


class CombinedLogger(NCA_Logger):
    def __init__(self, *loggers: NCA_Logger):
        super().__init__()
        self.loggers = [logger for logger in loggers if not isinstance(logger, NoOPLogger)]

    def init_training(self, trainer: "NCA_Trainer", stopping_criterion: "NCA_StoppingCriterion"):
        for logger in self.loggers:
            self.logger.info(f"Initializing logger {logger}")
            logger.init_training(trainer, stopping_criterion)

    def log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        for logger in self.loggers:
            logger.log(data_sample, state_progression, loss)

    def _log(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        raise RuntimeError("This method should never be called, as the .log()-Function is replaced")

    def close(self, data_sample: DataSample, state_progression: torch.Tensor, loss: LossResult):
        for logger in self.loggers:
            logger.close(data_sample, state_progression, loss)

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["_args_"] = [logger.to_config_dict() for logger in self.loggers]
        return cfg



from typing import Any, Optional

import torch
from jaxtyping import Float

from ncalib.trainer.loss_function import NamedLossFunction
from ncalib.utils.utils import slice_to_config


class OverflowLoss(NamedLossFunction):
    # https://colab.research.google.com/github/google-research/self-organising-systems/blob/master/isotropic_nca/blogpost_isonca_single_seed_pytorch.ipynb#scrollTo=3pxjdESnYlIC&line=1&uniqifier=1
    def __init__(
            self,
            channels: slice = slice(None, None, None),
            *,
            min: Optional[float] = None,
            max: Optional[float] = None,
            loss_name="OverflowLoss",
    ):
        super().__init__(loss_name=loss_name)
        self.channels = channels
        self.min = min
        self.max = max

    def _calculate_loss(
            self,
            state_progression: Float[torch.Tensor, "N B C W H"],
            target: Float[torch.Tensor, "B ..."],
            **kwargs
    ) -> Float[torch.Tensor, "B"]:
        clamped = torch.clamp(state_progression[:, :, self.channels], min=self.min, max=self.max)
        diff = state_progression[:, :, self.channels] - clamped
        return diff.square().sum(dim=[0, 2, 3, 4])

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["channels"] = slice_to_config(self.channels)
        cfg["min"] = self.min
        cfg["max"] = self.max
        return cfg


class DiffLoss(NamedLossFunction):
    # https://colab.research.google.com/github/google-research/self-organising-systems/blob/master/isotropic_nca/blogpost_isonca_single_seed_pytorch.ipynb#scrollTo=3pxjdESnYlIC&line=1&uniqifier=1
    def __init__(
            self,
            channels: slice = slice(None, None, None),
            *,
            loss_name="DiffLoss",
    ):
        super().__init__(loss_name)
        self.channels = channels

    def _calculate_loss(
            self,
            state_progression: Float[torch.Tensor, "N B C W H"],
            target: Float[torch.Tensor, "B ..."],
            **kwargs
    ) -> Float[torch.Tensor, "B"]:
        diffs = state_progression[1:, self.channels] - state_progression[:-1, self.channels]
        return diffs.abs().mean(dim=[2, 3, 4]).sum(dim=0)

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["channels"] = slice_to_config(self.channels)
        return cfg



import abc
from typing import Any, Iterable

import torch

from ncalib.trainer.dataset import DataSample
from ncalib.utils.utils import NCA_Base


class PoolEvictionPolicy(NCA_Base, abc.ABC):
    def __init__(self, weight: float = 1):
        super().__init__()
        self.weight = weight

    def __call__(self, data_sample: DataSample, losses: torch.Tensor):
        return self.weight * self.score_batch(data_sample, losses)

    def __add__(self, other: "PoolEvictionPolicy") -> "CombinedPoolEvictionPolicy":
        # Prepare each policy so that each is an iterable of simple policies
        if isinstance(self, CombinedPoolEvictionPolicy):
            policies1 = tuple(self.policies)
        else:
            policies1 = (self,)

        if isinstance(other, CombinedPoolEvictionPolicy):
            policies2 = tuple(other.policies)
        else:
            policies2 = (other,)

        # Combine Loss function
        return CombinedPoolEvictionPolicy(policies1 + policies2)

    def __mul__(self, value: float) -> "PoolEvictionPolicy":
        self.weight *= value
        return self

    def __rmul__(self, value) -> "PoolEvictionPolicy":
        return self.__mul__(value)

    @abc.abstractmethod
    def score_batch(self, data_sample: DataSample, losses: torch.Tensor) -> torch.Tensor:
        """
        Calculates scores between 0 and 1 for each sample
        :param data_sample:
        :param losses:
        :return:
        """
        pass

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["weight"] = self.weight
        return cfg


class CombinedPoolEvictionPolicy(PoolEvictionPolicy):
    def __init__(self, policies: Iterable[PoolEvictionPolicy], weight=1):
        super().__init__(weight=weight)
        self.policies = tuple(policies)

    def score_batch(self, data_sample: DataSample, losses: torch.Tensor) -> torch.Tensor:
        result = self.policies[0](data_sample, losses)
        for policy in self.policies[1:]:
            result += policy(data_sample, losses)

        return result

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["policies"] = [policy.to_config_dict() for policy in self.policies]
        return cfg



from pathlib import Path
from typing import Optional, Any

import torch

from ncalib.trainer.target import NCA_Target
from ncalib.utils.color_space import RGBA, ColorClass
from ncalib.utils.defaults import DEFAULT_DEVICE
from ncalib.utils.image import RGBAImage


class TensorTarget(NCA_Target):
    def __init__(
            self,
            image: RGBAImage,
            *,
            device: torch.device | str = DEFAULT_DEVICE,
            source: Optional[str] = None,
            color_space: ColorClass = RGBA(True, True, True)
    ):
        super().__init__(device=device)
        if not isinstance(image, RGBAImage):
            raise TypeError(f"Expected RGBAImage. Got {type(image)} instead.")
        self.image: RGBAImage = image
        self.source: Optional[str] = source
        self.color_space = color_space

        self._target: Optional[torch.Tensor] = None

    def __call__(self, n=1) -> torch.Tensor:
        """ :return Tensor [n x C x W x H] """
        return self.target.repeat([n, 1, 1, 1])

    def __repr__(self) -> str:
        if self.source is None:
            return f"{self.__class__.__name__}({self.channels}x{self.width}x{self.height})"
        return f"{self.__class__.__name__}({self.channels}x{self.width}x{self.height}, source={self.source})"

    @property
    def target(self) -> torch.Tensor:
        if self._target is None:
            self._target: torch.Tensor = self.color_space(self.image).to(self.device)
        return self._target

    @property
    def channels(self) -> int:
        return self.target.shape[0]

    @property
    def width(self) -> int:
        return self.image.width

    @property
    def height(self) -> int:
        return self.image.height

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["color_space"] = self.color_space.to_config_dict()

        return cfg

    @classmethod
    def load_from_path(
            cls,
            *,
            path: Path | str,
            width: int = None,
            height: int = None,
            padding: int = 0,
            color_space: Optional[ColorClass] = None,  # Default RGBA, if alpha is used, otherwise RGB
            device: str | torch.device = DEFAULT_DEVICE,
    ) -> NCA_Target:
        # Load target
        path = Path(path)
        image = RGBAImage.from_path(path)

        # Use default colorspace=RGBA, if alpha channel is used, otherwise RGB
        if color_space is None:
            use_alpha = not torch.all(image.rgba_alpha == 1)  # If all values are 1, we can omit the alpha channel
            color_space = RGBA(True, True, True, alpha=use_alpha)

        # Resize
        image = image.resize(width, height)

        # Padding
        if padding > 0:
            image = image.pad(padding)

        target = TensorTarget(
            image=image,
            color_space=color_space,
            device=device,
            source=str(path)
        )
        return target



import abc
from typing import Optional, Any

import torch
from tensordict import tensorclass, TensorDict
from torch.utils.data import IterableDataset

from ncalib.trainer.loss_function import LossResult
from ncalib.utils.utils import NCA_Base, make_circle_masks


@tensorclass
class DataSample:
    states: torch.Tensor
    targets: torch.Tensor
    n_steps: torch.Tensor
    training_regenerate: bool

    def get_kwargs(self, required_kwargs) -> TensorDict:
        result = {}
        for kwarg in required_kwargs:
            if kwarg == "original_state":
                # This is provided by nca itself
                continue
            try:
                result[kwarg] = getattr(self, kwarg)
            except KeyError:
                raise KeyError(f"Could not find {kwarg} in {self.__class__.__name__} but is required kwarg!")

        return TensorDict(result, batch_size=self.batch_size)

    def update_sample(self, new_state: torch.Tensor, loss: LossResult):
        self.states = new_state.detach()
        # if self.training_regenerate:
        #     B, C, W, H = new_state.shape
        #     circle_mask = make_circle_masks(B, W, H)
        #     self.states *= (1 - circle_mask.to(new_state.device))[:, None]

    @classmethod
    def merge_batch(cls, samples: list["DataSample"]) -> "DataSample":
        return torch.cat(samples, dim=0)


class NCA_Dataset(NCA_Base, IterableDataset, abc.ABC):
    def __init__(
            self,
            *,
            seed: Optional[int] = None,
    ):
        super().__init__()

        # rng is always for cpu
        self.rng = torch.Generator(device="cpu")
        if seed is not None:
            self.rng.manual_seed(seed)
        else:
            self.rng.seed()

    def __iter__(self):
        while True:
            yield self()

    @abc.abstractmethod
    def __call__(self) -> DataSample:
        """
        :return:
        DataSample consisting of at least:
            - state: Tensor[B x C x W x H]
            - target: Tensor[B x C' x W x H]
            - number_of_steps: Tensor[B]
        """
        pass


class NCA_SteppedDataset(NCA_Dataset):
    def __init__(self, step_range: tuple[int, int], *, seed: Optional[int]):
        super().__init__(seed=seed)
        self.step_range_min, self.step_range_max = step_range

    def generate_n_steps(self):
        return torch.randint(
            low=self.step_range_min,
            high=self.step_range_max + 1,
            size=(1,),
            generator=self.rng,
        )

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["step_range"] = (self.step_range_min, self.step_range_max)
        return cfg



import logging
from typing import Type, Any, Optional

import torch

from ncalib import NCA, NCA_Perception, ChainedPerception
from ncalib.nca.models import NCA_Model
from ncalib.nca.models.linear import LinearNCAModel
from ncalib.nca.perception.filter import IdentityPerception, SobelXPerception, SobelYPerception
from ncalib.nca.state_updater import NCA_StateUpdater
from ncalib.nca.state_updater.asynchronous_update import AsynchronousUpdate
from ncalib.nca.state_updater.simple_update import SimpleStateUpdater


class ReproducibilityError(Exception):
    pass


class ModularNCA(NCA):
    def __init__(
            self,
            channels: int,
            perception: NCA_Perception = ChainedPerception(perceptions=[
                IdentityPerception, SobelXPerception, SobelYPerception
            ]),
            model: NCA_Model | Type[NCA_Model] = LinearNCAModel,
            *,
            update_function: NCA_StateUpdater = AsynchronousUpdate() + SimpleStateUpdater(),
            device: torch.device | str = torch.device('cpu'),
            check_reproducibility=True,
    ):
        super().__init__(device=device)

        self.logger = logging.getLogger(self.__class__.__name__)
        # NCA Properties
        self._channels = channels

        # = Perception =
        self.perception = perception

        # = Model =
        if isinstance(model, type):
            if not issubclass(model, NCA_Model):
                raise TypeError(f"Expected type {NCA_Model} for model. Got {model}")
            self.model = model(self.perception.output_size_for_n_channels(channels), channels)
        else:
            self.model = model

        # = Post-processing =
        self.state_update_function = update_function

        # Update device
        super().to(device)
        self.model = self.model.to(self.device)
        # Perception and Post-processing gets device from state

        # Reproducibility check
        if check_reproducibility:
            self.reproducibility_check()

        # Properties
        self._last_perception: Optional[torch.Tensor] = None

    def compile(self, *, perception=True, model=True, update_function=True):
        if perception:
            self.perception = torch.compile(self.perception, mode="reduce-overhead")
        if model:
            self.model = torch.compile(self.model)  # Cannot use "reduce-overhead"
        if update_function:
            self.state_update_function = torch.compile(self.state_update_function, mode="reduce-overhead")

    def forward(self, state: torch.Tensor, **kwargs) -> torch.Tensor:
        # Perception
        perception = self.perception(state, **kwargs)

        # Model
        delta_state = self.model(perception, **kwargs)

        # Post-processing
        next_state, delta_state = self.state_update_function(
            state,
            delta_state,
            original_state=state.detach().clone(),
            perception=perception,
            **kwargs
        )

        self._last_perception = perception
        self._last_delta_state = delta_state
        return next_state

    @property
    def channels(self) -> int:
        return self._channels

    def reset(self):
        self.model.reset()
        self.perception.reset()
        self.state_update_function.reset()

    def reproducibility_check(self):
        config_dict = self.as_full_config_dict()
        nca = self.load_from(config_dict)

        if self.num_params != nca.num_params:
            raise ReproducibilityError("Number of parameters does not match!")

        self.logger.info("Reproducibility check passed!")

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["perception"] = self.perception.to_config_dict()
        cfg["model"] = self.model.to_config_dict()
        cfg["update_function"] = self.state_update_function.to_config_dict()
        cfg["channels"] = self.channels
        cfg["check_reproducibility"] = False  # Prevent infinite loop

        return cfg

    def to(
            self,
            device: torch.device | str
    ):
        super().to(device=device)
        self.model.to(device)

    def required_kwargs(self) -> set[str]:
        res = super().required_kwargs()
        res.update(self.perception.required_kwargs())
        res.update(self.model.required_kwargs())
        res.update(self.state_update_function.required_kwargs())
        return res



import abc
from typing import Callable, Literal

import cv2
import numpy as np
import torch


class GuiBrush(abc.ABC):
    def __init__(
            self,
            *,
            value: float | Callable[[int, int, int], np.ndarray | float] | Literal["random"] = 1,
            channels: slice = slice(None, None, None)
    ):
        super().__init__()
        if isinstance(value, float | int):
            self.value_factory = lambda c, w, h: value
        elif value == "random":
            self.value_factory = lambda c, w, h: torch.rand((c, w, h))
        else:
            self.value_factory = value
        self.channels = channels

    def modify_state(self, state: torch.Tensor, mouse: "MouseState") -> torch.Tensor:
        # Mouse event not parsed
        if mouse.batch is None:
            return state

        C, W, H = state[mouse.batch, self.channels].shape
        mask = np.zeros((W, H))

        if mouse.left_down:
            self._draw(mask, mouse)

        mask = torch.from_numpy(mask).to(state.device)
        state[mouse.batch, self.channels] = (
                state[mouse.batch, self.channels] * (mask == 0) +
                self.value_factory(C, W, H) * (mask != 0)
        )
        return state

    @abc.abstractmethod
    def _draw(self, mask: np.ndarray, mouse: "MouseState"):
        pass


class CircleBrush(GuiBrush):
    def __init__(self, *, value: float = 1, size=5, channels: slice = slice(None, None, None)):
        super().__init__(value=value, channels=channels)
        self.size = size

    def _draw(self, mask: np.ndarray, mouse: "MouseState"):
        cv2.circle(mask, (int(mouse.x), int(mouse.y)), self.size, 1, -1)



import time
import warnings
from typing import Optional, Any

import torch
from torch.utils.checkpoint import checkpoint_sequential

from ncalib.nca.modular_nca import ModularNCA
from ncalib.trainer import NCA_Trainer, NCA_StoppingCriterion, MaxEpochs, NCA_Logger
from ncalib.trainer.batcher import NCA_Batcher
from ncalib.trainer.logger import NoOPLogger
from ncalib.trainer.loss_function import NCA_LossFunction
from ncalib.utils.utils import optimizer_to_config, scheduler_to_config


class CombinedTrainer(NCA_Trainer):
    def __init__(
            self,
            nca: ModularNCA,
            data_batcher: NCA_Batcher,
            loss_function: NCA_LossFunction,
            *,
            use_checkpointing: bool = False,
            clip_grad_norm: Optional[float] = None,
            normalize_gradients: bool = True,
            optimizer: Optional[torch.optim.Optimizer] = None,
            scheduler: Optional[Any] = None,  # Needs to be of type torch.optim._LRScheduler
    ):
        super().__init__(
            nca=nca,
            optimizer=optimizer,
            scheduler=scheduler,
        )

        self.data_batcher = data_batcher
        self.loss_function = loss_function
        self.use_checkpointing = use_checkpointing
        self.clip_grad_norm = clip_grad_norm
        self.normalize_gradients = normalize_gradients

    def train(
            self,
            stopping_criterion: NCA_StoppingCriterion = MaxEpochs(1000),
            *,
            loggers: NCA_Logger | None = None
    ):
        if loggers is None:
            loggers = NoOPLogger()
        self.nca.train()
        self.data_batcher.init()

        self.logger.info(f"Initializing loggers...")
        loggers.init_training(self, stopping_criterion)

        sample = None
        loss = None
        state_progression = None

        self._time_started = time.perf_counter()
        self._epoch = 0
        while not stopping_criterion(self, sample, state_progression, loss):
            self._epoch += 1

            # Get state and target
            with torch.no_grad():
                sample = self.data_batcher.next_batch()
                state = sample.states.detach()
                kwargs = sample.get_kwargs(self.nca.required_kwargs())

            # Go through NCA
            # First use checkpointing for first steps if enabled
            # Step through remaining steps manually
            current_step = 0
            state_progression = [state.clone().detach()]
            if self.use_checkpointing:
                warnings.warn("Checkpoint currently seems broken. Please do NOT use!")
                if len(sample.kwargs) > 0:
                    raise NotImplementedError("NCA kwargs currently not implemented with checkpointing. "
                                              f"Please set `use_checkpointing` to `False`. "
                                              f"(Existing keys: {kwargs.keys()})")
                current_step = sample.n_steps.min()
                state = checkpoint_sequential(
                    [self.nca] * current_step,
                    16,  # Segments.
                    state
                )

            for step in range(current_step, sample.n_steps.max().item()):
                with torch.no_grad():
                    state_mask = (sample.n_steps >= step)
                    this_kwargs = {key: value[step] for key, value in kwargs.items()}

                state[state_mask] = self.nca(state[state_mask], **this_kwargs)
                state_progression.append(state.clone())
                # cv2.imshow("Step", state_to_bgr_numpy(state, channels=slice(1,4)))
                # cv2.imshow("Angle", state[0,0].detach().cpu().numpy())
                # cv2.waitKey(50)
            # cv2.waitKey(0)
            # cv2.destroyAllWindows()

            state_progression = torch.stack(state_progression, dim=0)

            # Calculate Loss
            loss = self.loss_function(state_progression, sample.targets)

            # Backpropagation
            total_loss = loss.total_loss(ignore_inf=False)

            if not torch.isfinite(total_loss):
                self.logger.warn(f"Ignoring loss! {loss.loss_tensor.mean().item()}")
            else:
                self.optimizer.zero_grad(set_to_none=True)
                self.fabric.backward(total_loss)

                if self.normalize_gradients:
                    for p in self.nca.parameters():
                        p.grad /= p.grad.norm() + 1e-8

                if self.clip_grad_norm is not None:
                    torch.nn.utils.clip_grad_norm_(self.nca.parameters(), self.clip_grad_norm)

                self.optimizer.step()
                if self.scheduler is not None:
                    self.scheduler.step()

            sample.update_sample(new_state=state, loss=loss)
            self.data_batcher.feedback(sample, new_state=state, loss=loss)

            # Logging
            self.losses.append(total_loss.detach().cpu().item())
            loggers.log(sample, state_progression=state_progression, loss=loss)

        # Finish and clean up training
        loggers.close(sample, state_progression=state_progression, loss=loss)

        self._time_finished = time.perf_counter()
        self.logger.info(f"Training finished after {self.training_duration} seconds. "
                         f"Reason {stopping_criterion.message}")

    def to_config_dict(self) -> dict[str, Any]:
        cfg = super().to_config_dict()
        cfg["nca"] = self.nca.to_config_dict()
        cfg["optimizer"] = optimizer_to_config(self.optimizer)
        cfg["scheduler"] = scheduler_to_config(self.scheduler)
        cfg["data_batcher"] = self.data_batcher.to_config_dict()
        cfg["loss_function"] = self.loss_function.to_config_dict()
        cfg["use_checkpointing"] = self.use_checkpointing
        cfg["clip_grad_norm"] = self.clip_grad_norm
        cfg["normalize_gradients"] = self.normalize_gradients
        return cfg
